// SPDX-License-Identifier: 0BSD
/*!
 * @file
 * @license{
 * BSD Zero Clause License
 *
 * Copyright (c) 2022 by Raphael Bach
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 * }
 */
/*==============================================================================
    GUARD
==============================================================================*/
#ifndef FMPI_ERROR_H_20220510231118
#define FMPI_ERROR_H_20220510231118
/*==============================================================================
    INCLUDE
==============================================================================*/
// C Standard Library
#include <time.h> // size_t, time_t
/*==============================================================================
    DEFINE
==============================================================================*/
/**
 * Size needed for a buffer to contain the timestamp string returned by
 * fmpi_error_timestamp_str().
 */
#define FMPI_ERROR_STR_BUF_SIZE sizeof("YYYY-MM-DDThh:mm:ss+hhmm")
/**
 * Maximum length of an error message that can be passed to fmpi_raise_error()
 * or FMPI_RAISE_ERROR().
 */
#define FMPI_ERROR_MSG_LENGTH_MAX 4096
/*==============================================================================
    TYPE
==============================================================================*/
/*------------------------------------------------------------------------------
    fmpi_error_id
------------------------------------------------------------------------------*/
typedef enum fmpi_error_id {
    FMPI_ERROR = -1,
    FMPI_SUCCESS = 0,
    FMPI_ERR_MPI,
    FMPI_ERR_FUTHARK
} fmpi_error_id;
/*------------------------------------------------------------------------------
    fmpi_error
------------------------------------------------------------------------------*/
/**
 * Holds information about an error(what, where and when).
 *
 * @example{
 *  struct fmpi_error error = {
 *      .file = __FILE__,
 *      .line = __LINE__,
 *      .func = __func__,
 *      .time = time(NULL),
 *      .module = "MPI",
 *      .msg = "Something failed!"
 *  };
 * }
 */
typedef struct fmpi_error {
    const char * file;   //!< Name of the file where the error was raised.
    size_t line;         //!< Line number on which the the error was raised.
    const char * func;   //!< Name of the funcion in which the error was raised.
    time_t time;         //!< Time in seconds since Epoch.
    const char * module; //!< Module name(e.g FUTHARK or MPI).
    const char * msg;    //!< Message describing the error.
} fmpi_error;
/*------------------------------------------------------------------------------
    fmpi_err_func
------------------------------------------------------------------------------*/
/**
 * Function pointer type for error callbacks.
 *
 * @param[in] err       : A pointer to a fmpi_error structure.
 * @param[in] user_data : A pointer to any data needed in the error callback.
 *
 * @return Nothing.
 *
 * @warning
 * - \b [UB] \p{err} must not be `NULL`.
 *
 * @example{
 *  void error_callback(const struct fmpi_error * err, const void * user_data) {
 *      fprintf(stderr, "%s:%zu:%s() - Error: %s value=%d\n",
 *          err->file, err->line, err->func, err->msg, *(const int *)user_data
 *      );
 *  }
 *  int main(void) {
 *      const fmpi_err_func err_func = error_callback;
 *      const struct fmpi_error error = {
 *          .file = __FILE__,
 *          .line = __LINE__,
 *          .func = __func__,
 *          .time = time(NULL),
 *          .module = "MPI",
 *          .msg = "Something failed!"
 *      };
 *      const int special_value = 3;
 *      err_func(error, &special_value);
 *  }
 * }
 */
typedef void (*fmpi_err_func)(const struct fmpi_error * err, const void * user_data);
/*------------------------------------------------------------------------------
    fmpi_error_handler
------------------------------------------------------------------------------*/
/**
 * Holds an error callback and a pointer to data the user may wish to pass to
 * the callback.
 *
 * @example{
 *  void error_callback(const struct fmpi_error * err, const void * user_data) {
 *      fprintf(stderr, "%s:%zu:%s() - Error: %s value=%d\n",
 *          err->file, err->line, err->func, err->msg, *(const int *)user_data
 *      );
 *  }
 *  int main(void) {
 *      const int user_data = 3;
 *      const struct fmpi_error_handler err_handler = {
 *          .func = error_callback,
 *          .user_data = &user_data
 *      };
 *  }
 * }
 */
typedef struct fmpi_error_handler {
    fmpi_err_func func; //!< Error callback.
    void * user_data;   //!< Pointer to user data.
} fmpi_error_handler;
/*==============================================================================
    FUNCTION
==============================================================================*/
/*------------------------------------------------------------------------------
    fmpi_raise_error()
------------------------------------------------------------------------------*/
/**
 * Raises an error with the given informations and trigger the callback if there
 * is one.
 *
 * This function constucts a fmpi_error structure with its arguments and calls
 * the error callback stored in \p{err_handler} if they both are not `NULL`.
 * The `fmpi_error` struct and the `user_data` member of \p{err_handler} are
 * passed to the callback.
 *
 * - Returns immediately if \p{err_handler}`->func()` is `NULL`.
 * - The last two parameters are to be used in the same way as in the `printf()`
 *   family functions.
 * - The `time` member of the fmpi_error structure is set to `time(NULL)` at the
 *   beginning of the function.
 * - The `msg` member of the fmpi_error structure is set to
 *   `<Error getting message>` if there is an error parsing \p{fmt} or \p{...}.
 * - This function is mainly used by macros which automatically set most of the
 *   parameters.
 *
 * @param[in] err_handler : Pointer to an error handler.
 * @param[in] file        : Name of the file where the error was raised
 * @param[in] line        : Line number on which the the error was raised.
 * @param[in] func        : Name of the funcion in which the error was raised.
 * @param[in] module      : Name of the module in which the error was raised.
 * @param[in] fmt         : Null-terminated format string.
 * @param[in] ...         : List of arguments for \p{fmt}.
 *
 * @return Nothing.
 *
 * @warning
 * - \b [UB] \p{err_handler} must not be `NULL`.
 * - \b [UB] \p{file} must not be `NULL`.
 * - \b [UB] \p{func} must not be `NULL`.
 * - \b [UB] \p{module} must not be `NULL`.
 * - \b [UB] \p{fmt} must not be `NULL`.
 * - \b [UB] The final string length after formatting(including the terminating
 *   null character) of \p{fmt} and \p{...} must be <= #FMPI_ERROR_MSG_LENGTH_MAX.
 *
 * @example{
 *  void error_callback(const struct fmpi_error * err, const void * user_data) {
 *      fprintf(stderr, "%s:%zu:%s() - Error: %s value=%d\n",
 *          err->file, err->line, err->func, err->msg, *(const int *)user_data
 *      );
 *  }
 *  int main(void) {
 *      const int user_data = 3;
 *      const struct fmpi_error_handler err_handler = {
 *          .func = error_callback,
 *          .user_data = &user_data
 *      };
 *      fmpi_raise_error(
 *          &err_handler, __FILE__, __LINE__, __func__, "FMPI", "Something failed!"
 *      );
 *  }
 * }
 */
void fmpi_raise_error(
    struct fmpi_error_handler err_handler,
    const char * file, size_t line, const char * func, const char * module,
    const char * fmt, ...
);
/*------------------------------------------------------------------------------
    fmpi_error_timestamp_str()
------------------------------------------------------------------------------*/
/**
 * Converts time in seconds since Epoch to a null-terminated ISO 8601 formatted
 * string.
 *
 * - Example of an ISO 8601 time string: `2017-08-23T14:46:59+0430`.
 *
 * @param[in]  t    : Time in seconds since Epoch.
 * @param[out] buf  : Buffer where the string will be stored.
 * @param[in]  size : Size of \p{buf}.
 *
 * @return
 * - @success: A pointer to \p{buf}.
 * - @failure: `NULL`.
 *
 * @warning
 * - \b [UB] \p{buf} must not be `NULL`.
 * - \b [UB] \p{size} must be >= #FMPI_ERROR_STR_BUF_SIZE.
 *
 * @example{
 *  char timestamp_buf[FMPI_ERROR_STR_BUF_SIZE] = {'\0'};
 *  const char * timestamp_str = timestamp_buf;
 *  const char * ret = fmpi_error_timestamp_str(err->time, timestamp_buf, sizeof(timestamp_buf));
 *  if(ret == NULL) {
 *      timestamp_str = "<Error getting timestamp>";
 *  }
 * }
 */
const char * fmpi_error_timestamp_str(time_t t, char * buf, size_t size);
/*==============================================================================
    MACRO
==============================================================================*/
/*------------------------------------------------------------------------------
    FMPI_RAISE_ERROR()
------------------------------------------------------------------------------*/
/**
 * Macro easing the call to fmpi_raise_error() by providing the `file`, `line`
 * and `func` argument.
 *
 * @param[in] err_handler : Pointer to an error handler.
 * @param[in] module      : Name of the module in which the error was raised.
 * @param[in] ...         : Null-terminated format string for the first
 *                          parameter and the arguments for the format string
 *                          itself after.
 *
 * @return Nothing.
 *
 * @warning
 * - \b [UB] \p{err_handler} must not be `NULL`.
 * - \b [UB] \p{module} must not be `NULL`.
 * - \b [UB] \p{fmt} must not be `NULL`.
 * - \b [UB] The final string length after formatting(including the terminating
 *   null character) of \p{...} must be <= #FMPI_ERROR_MSG_LENGTH_MAX.
 *
 * @example{
 *  void error_callback(const struct fmpi_error * err, const void * user_data) {
 *      fprintf(stderr, "%s:%zu:%s() - Error: %s value=%d\n",
 *          err->file, err->line, err->func, err->msg, *(const int *)user_data
 *      );
 *  }
 *  int main(void) {
 *      const int user_data = 3;
 *      const struct fmpi_error_handler err_handler = {
 *          .func = error_callback,
 *          .user_data = &user_data
 *      };
 *      FMPI_RAISE_ERROR(err_handler, "FMPI", "Error: %s", "something failed!\n");
 *  }
 * }
 */
#define FMPI_RAISE_ERROR(err_handler, module, ...) \
    fmpi_raise_error(err_handler, __FILE__, __LINE__, __func__, (module), __VA_ARGS__)
/*==============================================================================
    GUARD
==============================================================================*/
#endif // FMPI_ERROR_H_20220510231118
