// SPDX-License-Identifier: 0BSD
/*!
 * @file
 * @license{
 * BSD Zero Clause License
 *
 * Copyright (c) 2022 by Raphael Bach
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 * }
 */
/*==============================================================================
    GUARD
==============================================================================*/
#ifndef FMPI_GENERIC_H_20211201184118
#define FMPI_GENERIC_H_20211201184118
/*==============================================================================
    INCLUDE
==============================================================================*/
// External
#include "../../../external/cpl/cpl_map.h"
#include "../../../external/cpl/cpl_token.h"
#include "../../../external/cpl/cpl_util.h"
/*==============================================================================
    MACRO
==============================================================================*/
/*------------------------------------------------------------------------------
    FMPI_GENERIC_FUNC()
------------------------------------------------------------------------------*/
/**
 * Expands to a `_Generic()` expression.
 *
 * The `generic-association` are made according to the following syntax:
 * `<type> : fmpi_<func>_<type>`.
 *
 * - `<func>` is the \p{FUNC} parameter and `<type>` is a type from the \p{...}
 *   parameter.
 *
 * @param CTR_EXPR : Controlling expression
 * @param FUNC     : Main part of the function name(must be a single word).
 * @param ...      : Comma separated list of type names(each must be a single word).
 *
 * @expansion{
 *  FMPI_GENERIC_FUNC(x, foo, char, int) =>
 *      _Generic((x),
 *          default : 0,
 *          char : fmpi_foo_char,
 *          int  : fmpi_foo_int
 *      )
 * }
 *
 * @example{
 * int x = FMPI_GENERIC_FUNC(x, foo, char, int)(a, b); // x = fmpi_foo_int(a, b)
 * }
 */
#define FMPI_GENERIC_FUNC(CTR_EXPR, FUNC, ...) \
    _Generic((*CTR_EXPR), default: 0, \
        CPL_MAP_FIXED(FMPI_GENERIC_ASSOC, CPL_COMMA, (FUNC), __VA_ARGS__) \
    )
/*------------------------------------------------------------------------------
    FMPI_GENERIC_ASSOC()
------------------------------------------------------------------------------*/
/**
 * Expands to a `generic-association` according to the following pattern:
 * `<TYPE> : fmpi_<FUNC>_<TYPE>`.
 *
 * - Only useful when used in the #FMPI_GENERIC_FUNC macro.
 *
 * @param FUNC : Main part of the function name(must be a single word).
 * @param TYPE : The name of a type(must be a single word).
 *
 * @expansion{
 *  FMPI_GENERIC_ASSOC(foo, int) =>
 *      int : fmpi_foo_int
 * }
 */
#define FMPI_GENERIC_ASSOC(FUNC, TYPE) \
    TYPE : fmpi_##FUNC##_##TYPE
/*------------------------------------------------------------------------------
    FMPI_DECLARE_FUNCS_T()
------------------------------------------------------------------------------*/
/**
 * Declares one or more functions.
 *
 * Applies \p{M()} to each argument of \p{...} and add a semicolon `;` between
 * each call.
 *
 * @param M   : A macro having a `M(T)` signature.
 * @param ... : Comma separated list of type names (each must be a single word).
 *
 * @expansion{
 *  FMPI_DECLARE_FUNCS_T(M, char, int, double) =>
 *      M(char) ; M(int) ; M(double)
 * }
 *
 * @example{
 *  #define M(T) T foo_##T(T x); T bar_##T(T x)
 *  FMPI_DECLARE_FUNCS_T(M, char, int);
 *  // char foo_char(char x);
 *  // char bar_char(char x);
 *  // int foo_int(int x);
 *  // int bar_int(int x);
 * }
 */
#define FMPI_DECLARE_FUNCS_T(M, ...) \
    CPL_MAP(M, CPL_SEMICOLON, __VA_ARGS__)
/*------------------------------------------------------------------------------
    FMPI_DEFINE_FUNCS_T()
------------------------------------------------------------------------------*/
/**
 * Defines one or more functions.
 *
 * Applies \p{M()} to each argument of \p{...}.
 *
 * @param M   : A macro having a `M(T)` signature.
 * @param ... : Comma separated list of type names (each must be a single word).
 *
 * @expansion{
 *  FMPI_DEFINE_FUNCS_T(M, char, int, double) =>
 *      M(char) M(int) M(double)
 * }
 *
 * @example{
 *  #define M(T) \
 *  T foo_##T(T x) { \
 *      return bar_##T(x); \
 *  } \
 *  T bar_##T(T x) { \
 *      return x; \
 *  }
 *  FMPI_DEFINE_FUNCS_T(M, char, int)
 *  // char foo_char(char x) { return bar_char(x); }
 *  // char bar_char(char x) { return x; }
 *  // int foo_int(int x) { return bar_int(x); }
 *  // int bar_int(int x) { return x; }
 * }
 */
#define FMPI_DEFINE_FUNCS_T(M, ...) \
    CPL_MAP(M, CPL_EMPTY, __VA_ARGS__)
/*------------------------------------------------------------------------------
    FMPI_DECLARE_FUNCS_DT()
------------------------------------------------------------------------------*/
/**
 * TODO
 *
 * @param M   : A macro having a `M(T, D)` signature.
 * @param D   :
 * @param ... : Comma separated list of type names (each must be a single word).
 *
 * @expansion{
 *  TODO
 * }
 *
 * @example{
 *  TODO
 * }
 */
#define FMPI_DECLARE_FUNCS_DT(M, D, ...) \
    CPL_MAP_FIXED(M, CPL_SEMICOLON, (D), __VA_ARGS__)
/*------------------------------------------------------------------------------
    FMPI_DEFINE_FUNCS_DT()
------------------------------------------------------------------------------*/
/**
 * TODO
 *
 * @param M   : A macro having a `M(T, D)` signature.
 * @param D   :
 * @param ... : Comma separated list of type names (each must be a single word).
 *
 * @expansion{
 *  TODO
 * }
 *
 * @example{
 *  TODO
 * }
 */
#define FMPI_DEFINE_FUNCS_DT(M, D, ...) \
    CPL_MAP_FIXED(M, CPL_EMPTY, (D), __VA_ARGS__)
/*==============================================================================
    GUARD
==============================================================================*/
#endif // FMPI_GENERIC_H_20211201184118
