// SPDX-License-Identifier: 0BSD
/*!
 * @file
 * @license{
 * BSD Zero Clause License
 *
 * Copyright (c) 2022 by Raphael Bach
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 * }
 */
/*==============================================================================
    INCLUDE
==============================================================================*/
// Own header
#include "internal/fmpi_ctx.h"
// C Standard Library
#include <assert.h>
#include <stdio.h>  // fprintf()
#include <stdlib.h> // NULL, free(), malloc()
// Internal
#include "internal/fmpi_error.h"
#include "internal/fmpi_futhark.h"
#include "internal/fmpi_mpi.h"
/*==============================================================================
    PRIVATE FUNCTION DECLARATION
==============================================================================*/
/*------------------------------------------------------------------------------
    fmpi_mpi_error_callback()
------------------------------------------------------------------------------*/
static void fmpi_error_callback(
    const struct fmpi_error * err, const void * user_data
);
/*------------------------------------------------------------------------------
    fmpi_ctx_print_error()
------------------------------------------------------------------------------*/
static void fmpi_ctx_print_error(
    const struct fmpi_error * err, const void * user_data
);
/*==============================================================================
    PUBLIC FUNCTION DEFINITION
==============================================================================*/
/*------------------------------------------------------------------------------
    fmpi_ctx_create()
------------------------------------------------------------------------------*/
struct fmpi_ctx * fmpi_ctx_create(int * const argc, char ** argv[])
{
    struct fmpi_ctx * ctx = malloc(sizeof(*ctx));
    struct fmpi_error_handler err_handler = {
        .func = fmpi_error_callback,
        .user_data = NULL
    };
    if(ctx == NULL) {
        //! @todo Find if there is a way to get informations about the
        //! environment before MPI is initialized.
        FMPI_RAISE_ERROR(err_handler, "FMPI", "malloc(fmpi_ctx) failed!");
        return NULL;
    }
    err_handler.user_data = ctx;
    ctx->err_handler = err_handler;
    ctx->task_cnt = 0;
    // MPI initialization
    ctx->mpi = fmpi_mpi_init(argc, argv, (struct fmpi_error_handler) {
        .func = fmpi_error_callback,
        .user_data = ctx
    });
    if(ctx->mpi == NULL) {
        FMPI_RAISE_ERROR(err_handler, "FMPI", "fmpi_mpi_init() failed!");
        if(fmpi_ctx_destroy(&ctx) != FMPI_SUCCESS) {
            FMPI_RAISE_ERROR(err_handler, "FMPI", "fmpi_ctx_destroy() failed!");
        }
        return NULL;
    }
    // Futhark Initialization
    ctx->fut = fmpi_futhark_init((struct fmpi_error_handler) {
        .func = fmpi_error_callback,
        .user_data = ctx
    });
    if(ctx->fut == NULL) {
        FMPI_RAISE_ERROR(err_handler, "FMPI", "fmpi_futhark_init() failed!");
        if(fmpi_ctx_destroy(&ctx) != FMPI_SUCCESS) {
            FMPI_RAISE_ERROR(err_handler, "FMPI", "fmpi_ctx_destroy() failed!");
        }
        return NULL;
    }
    return ctx;
}
/*------------------------------------------------------------------------------
    fmpi_ctx_destroy()
------------------------------------------------------------------------------*/
int fmpi_ctx_destroy(struct fmpi_ctx ** const ctx)
{
    assert(ctx != NULL);
    assert(*ctx != NULL);

    int err = FMPI_SUCCESS;
    if((*ctx)->fut != NULL && fmpi_futhark_exit(&(*ctx)->fut) != FMPI_SUCCESS) {
        FMPI_RAISE_ERROR((*ctx)->err_handler, "FMPI", "fmpi_futhark_exit() failed!");
        err = FMPI_ERROR;
    }
    if((*ctx)->mpi != NULL && fmpi_mpi_exit(&(*ctx)->mpi) != FMPI_SUCCESS) {
        FMPI_RAISE_ERROR((*ctx)->err_handler, "FMPI", "fmpi_mpi_exit() failed!");
        err = FMPI_ERROR;
    }
    free(*ctx); *ctx = NULL;
    return err;
}
/*==============================================================================
    PRIVATE FUNCTION DEFINITION
==============================================================================*/
/*------------------------------------------------------------------------------
    fmpi_error_callback()
------------------------------------------------------------------------------*/
static void fmpi_error_callback(
    const struct fmpi_error * const err, const void * const user_data
){
    assert(err != NULL);
    fmpi_ctx_print_error(err, user_data);
}
/*------------------------------------------------------------------------------
    fmpi_ctx_print_error()
------------------------------------------------------------------------------*/
static void fmpi_ctx_print_error(
    const struct fmpi_error * const err, const void * const user_data
){
    assert(err != NULL);
    int rank = -1;
    if(user_data != NULL) {
        rank = ((const fmpi_ctx *)user_data)->mpi->rank;
    }
    char timestamp_buf[FMPI_ERROR_STR_BUF_SIZE] = {'\0'};
    const char * timestamp_str = timestamp_buf;
    const char * ret = fmpi_error_timestamp_str(err->time, timestamp_buf, sizeof(timestamp_buf));
    if(ret == NULL) {
        timestamp_str = "<Error getting timestamp>";
    }
    fprintf(stderr, "%s - %s:%zu:%s() - [%s ERROR rank=%d] %s\n",
        timestamp_str, err->file, err->line, err->func, err->module, rank, err->msg
    );
}
