// SPDX-License-Identifier: 0BSD
/*!
 * @file
 * @license{
 * BSD Zero Clause License
 *
 * Copyright (c) 2022 by Raphael Bach
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 * }
 */
/*==============================================================================
    INCLUDE
==============================================================================*/
// Own header
#include "fmpi_domain.h"
// C Standard Library
#include <assert.h>
#include <limits.h> // INT_MAX
#include <stdlib.h> // size_t, NULL, malloc(), free()
#include <string.h> // memcpy()
// Internal
#include "fmpi_data.h"
#include "fmpi_stencil.h"
#include "internal/fmpi_ctx.h"
#include "internal/fmpi_error.h"
#include "internal/fmpi_mpi.h"
/*==============================================================================
    PRIVATE FUNCTION DECLARATION
==============================================================================*/
/*------------------------------------------------------------------------------
    fmpi_partition_block_1d()
------------------------------------------------------------------------------*/
/**
 * TODO
 */
static struct fmpi_data fmpi_partition_block_1d(
    const struct fmpi_ctx * ctx, const struct fmpi_data * data
);
/*------------------------------------------------------------------------------
    fmpi_partition_block_2d()
------------------------------------------------------------------------------*/
static struct fmpi_data fmpi_partition_block_2d(
    const struct fmpi_ctx * ctx, const struct fmpi_data * data
);
/*------------------------------------------------------------------------------
    fmpi_halo_1d()
------------------------------------------------------------------------------*/
static struct fmpi_data fmpi_halo_1d(
    const struct fmpi_ctx * ctx, struct fmpi_data data,
    struct fmpi_stencil stencil
);
/*------------------------------------------------------------------------------
    fmpi_halo_2d()
------------------------------------------------------------------------------*/
static struct fmpi_data fmpi_halo_2d(
    const struct fmpi_ctx * ctx, struct fmpi_data data,
    struct fmpi_stencil stencil
);
/*==============================================================================
    PUBLIC FUNCTION DEFINITION
==============================================================================*/
//! @todo Handle creation of domain with data in 2D and 3D
/*------------------------------------------------------------------------------
    fmpi_new_domain()
------------------------------------------------------------------------------*/
struct fmpi_domain fmpi_new_domain(
    const struct fmpi_ctx * const ctx, const struct fmpi_data * const data,
    const struct fmpi_stencil stencil
){
    assert(ctx != NULL);
    const size_t proc_cnt = (size_t)ctx->mpi->size;
    struct fmpi_domain domain = {
        .data = data,
        .stencil = stencil,
        .part_cnt = proc_cnt
    };
    switch(data->dim_cnt) {
        case 1: {
            domain.inner = fmpi_partition_block_1d(ctx, data);
            break;
        }
        case 2: {
            domain.inner = fmpi_partition_block_2d(ctx, data);
            break;
        }
        default: break;
    }
    if(stencil.type != FMPI_STENCIL_NONE) {
        switch(data->dim_cnt) {
            case 1: {
                domain.halo = fmpi_halo_1d(ctx, domain.inner, stencil);
                domain.inner.raw = (char *)domain.halo.raw + (data->type.size * stencil.length);
                break;
            }
            case 2: {
                domain.halo = fmpi_halo_2d(ctx, domain.inner, stencil);
                const size_t offset = stencil.length * (domain.halo.dim_len[0] + 1) * data->type.size;
                domain.inner.raw = (char *)domain.halo.raw + offset;
                break;
            }
            default: break;
        }
    }
    return domain;
}
/*------------------------------------------------------------------------------
    fmpi_domain_set_inner()
------------------------------------------------------------------------------*/
void fmpi_domain_set_inner(
    const struct fmpi_ctx * const ctx, const struct fmpi_domain * const domain,
    const void * const data
){
    assert(ctx != NULL);
    assert(domain != NULL);
    assert(data != NULL);
    const size_t dim_cnt = domain->inner.dim_cnt;
    if(dim_cnt == 1) {
        memcpy(domain->inner.raw, data, domain->inner.size);
    } else if(dim_cnt == 2) {
        const size_t type_size = domain->inner.type.size;
        const size_t dim_len_y = domain->inner.dim_len[1];
        const size_t src_size_x =  domain->inner.dim_len[0] * type_size;
        for(size_t i = 0; i < dim_len_y; i++) {
            const size_t dest_offset = i * domain->halo.dim_len[0] * type_size;
            void * const dest = (char *)domain->inner.raw + dest_offset;
            const void * const src = (const char *)data + (i * src_size_x);
            memcpy(dest, src, src_size_x);
        }
    }
}
/*------------------------------------------------------------------------------
    fmpi_halo_exchange_1d()
------------------------------------------------------------------------------*/
int fmpi_halo_exchange_1d(
    const struct fmpi_ctx * const ctx, const struct fmpi_domain * const domain
){
    assert(ctx != NULL);
    assert(domain != NULL);

    const size_t idx_1 = domain->halo.type.size;
    const size_t idx_nm2 = (domain->halo.cnt-2) * domain->halo.type.size;
    const size_t idx_nm1 = idx_nm2 + idx_1;

          void * const buf_0   = domain->halo.raw;
    const void * const buf_1   = (char *)buf_0 + idx_1;
    const void * const buf_nm2 = (char *)buf_0 + idx_nm2;
          void * const buf_nm1 = (char *)buf_0 + idx_nm1;

    const int rank = ctx->mpi->rank;
    const int left = (rank != 0) ? (rank - 1) : MPI_PROC_NULL;
    const int right = (rank != (ctx->mpi->size-1)) ? (rank + 1) : MPI_PROC_NULL;
    MPI_Datatype type = fmpi_mpi_type(domain->halo.type.base);
    MPI_Sendrecv(
        buf_nm2, 1, type, right, 0,
        buf_0  , 1, type, left , 0,
        ctx->mpi->world, MPI_STATUS_IGNORE
    );
    MPI_Sendrecv(
        buf_1  , 1, type, left , 1,
        buf_nm1, 1, type, right, 1,
        ctx->mpi->world, MPI_STATUS_IGNORE
    );
    return FMPI_SUCCESS;
}
/*------------------------------------------------------------------------------
    fmpi_halo_exchange_2d()
------------------------------------------------------------------------------*/
int fmpi_halo_exchange_2d(
    const struct fmpi_ctx * const ctx, const struct fmpi_domain * const domain
){
    assert(ctx != NULL);
    assert(domain != NULL);

    const size_t type_size = domain->inner.type.size;
          void * const halo_top     = domain->halo.raw;
    const void * const inner_top    = (const char *)halo_top  + (domain->halo.dim_len[0] * domain->stencil.length       * type_size);
    const void * const inner_bottom = (const char *)inner_top + (domain->halo.dim_len[0] * (domain->inner.dim_len[1]-1) * type_size);
          void * const halo_bottom  = (char *)halo_top        + (domain->halo.dim_len[0] * (domain->halo.dim_len [1]-1) * type_size);

    const int rank = ctx->mpi->rank;
    const int above = (rank != 0) ? (rank - 1) : MPI_PROC_NULL;
    const int under = (rank != (ctx->mpi->size-1)) ? (rank + 1) : MPI_PROC_NULL;
    assert(domain->inner.dim_len[0] <= INT_MAX);
    const int cnt = (int)domain->halo.dim_len[0];
    MPI_Datatype type = fmpi_mpi_type(domain->halo.type.base);
    MPI_Sendrecv(
        inner_top  , cnt, type, above, 0,
        halo_bottom, cnt, type, under, 0,
        ctx->mpi->world, MPI_STATUS_IGNORE
    );
    MPI_Sendrecv(
        inner_bottom, cnt, type, under, 1,
        halo_top    , cnt, type, above, 1,
        ctx->mpi->world, MPI_STATUS_IGNORE
    );
    return FMPI_SUCCESS;
}
/*==============================================================================
    PRIVATE FUNCTION DEFINITION
==============================================================================*/
/*------------------------------------------------------------------------------
    fmpi_partition_block_1d()
------------------------------------------------------------------------------*/
static struct fmpi_data fmpi_partition_block_1d(
    const struct fmpi_ctx * const ctx, const struct fmpi_data * const data
) {
    assert(ctx != NULL);
    assert(data != NULL);

    const size_t rank = (size_t)ctx->mpi->rank;
    const size_t cnt_per_proc = data->cnt/(size_t)ctx->mpi->size;
    const size_t rem = data->cnt % (size_t)ctx->mpi->size;
    const size_t cnt = (rank < rem) ? (cnt_per_proc + 1) : cnt_per_proc;
    const size_t size = cnt * data->type.size;
    const size_t offset = (rank * size) + (rank < rem ? 0 : (rem * data->type.size));
    return (struct fmpi_data){
        .type = data->type,
        .cnt = cnt,
        .size = size,
        .dim_len = {cnt, 1, 1},
        .dim_cnt = 1,
        .raw = (char *)data->raw + offset,
        .gpu = NULL
    };
}
/*------------------------------------------------------------------------------
    fmpi_partition_block_2d()
------------------------------------------------------------------------------*/
static struct fmpi_data fmpi_partition_block_2d(
    const struct fmpi_ctx * const ctx, const struct fmpi_data * const data
) {
    assert(ctx != NULL);
    assert(data != NULL);

    size_t dim_len[1] = {0};
    if(fmpi_mpi_dims_create(ctx->mpi, dim_len, 1) != FMPI_SUCCESS) {
        FMPI_RAISE_ERROR(ctx->err_handler, "FMPI",
            "fmpi_mpi_dims_create() failed!"
        );
    }
    const size_t rank = (size_t)ctx->mpi->rank;
    const size_t cnt_per_proc = data->cnt/dim_len[0];
    const size_t rem = data->cnt % dim_len[0];
    const size_t cnt = (rank < rem) ? (cnt_per_proc + 1) : cnt_per_proc;
    const size_t size = cnt * data->type.size;
    const size_t offset = (rank * size) + (rank < rem ? 0 : (rem * data->type.size));
    return (struct fmpi_data){
        .type = data->type,
        .cnt = cnt,
        .size = size,
        .dim_len = {data->dim_len[0], data->dim_len[1]/dim_len[0], 1},
        .dim_cnt = 2,
        .raw = (char *)data->raw + offset,
        .gpu = NULL
    };
}
/*------------------------------------------------------------------------------
    fmpi_halo_1d()
------------------------------------------------------------------------------*/
static struct fmpi_data fmpi_halo_1d(
    const struct fmpi_ctx * const ctx, const struct fmpi_data data,
    const struct fmpi_stencil stencil
){
    assert(ctx != NULL);

    const size_t cnt = data.cnt + (2 * stencil.length);
    const size_t type_size = data.type.size;
    const size_t size = cnt * type_size;
    struct fmpi_data halo = {
        .type = data.type,
        .cnt = cnt,
        .size = size,
        .dim_len = {cnt, 1, 1},
        .dim_cnt = 1,
        .raw = calloc(cnt, type_size),
        .gpu = NULL
    };
    if(halo.raw == NULL) {
        FMPI_RAISE_ERROR(ctx->err_handler, "FMPI",
            "calloc(halo.raw) failed!"
        );
    }
    const size_t rank = (size_t)ctx->mpi->rank;
    // First rank
    if(rank == 0) {
        memcpy(((char *)halo.raw + type_size), data.raw, (data.size + type_size));
    // Last rank
    } else if(rank == (size_t)(ctx->mpi->size - 1)) {
        memcpy(halo.raw, ((const char *)(data.raw) - type_size), (data.size + type_size));
    // Middle rank
    } else {
        memcpy(halo.raw, ((const char *)(data.raw) - type_size), size);
    }
    return halo;
}
/*------------------------------------------------------------------------------
    fmpi_halo_2d()
------------------------------------------------------------------------------*/
static struct fmpi_data fmpi_halo_2d(
    const struct fmpi_ctx * const ctx, const struct fmpi_data data,
    const struct fmpi_stencil stencil
){
    assert(ctx != NULL);

    const size_t dim_len_x = data.dim_len[0] + (2 * stencil.length);
    const size_t dim_len_y = data.dim_len[1] + (2 * stencil.length);
    const size_t cnt = dim_len_x * dim_len_y;
    const size_t type_size = data.type.size;
    const size_t size = cnt * type_size;
    struct fmpi_data halo = {
        .type = data.type,
        .cnt = cnt,
        .size = size,
        .dim_len = {dim_len_x, dim_len_y, 1},
        .dim_cnt = 2,
        .raw = calloc(cnt, type_size),
        .gpu = NULL
    };
    if(halo.raw == NULL) {
        FMPI_RAISE_ERROR(ctx->err_handler, "FMPI",
            "calloc(halo.raw) failed!"
        );
    }
    const size_t rank = (size_t)ctx->mpi->rank;
    // First rank
    if(rank == 0) {
        for(size_t i = 0; i < data.dim_len[1] + stencil.length; i++) {
            const size_t dest_offset = ((i + stencil.length) * dim_len_x + stencil.length) * type_size;
            void * const dest = (char *)halo.raw + dest_offset;
            const size_t len = data.dim_len[0] * type_size;
            const void * const src = (char *)data.raw + (i * len);
            memcpy(dest, src, len);
        }
    // Last rank
    } else if(rank == (size_t)(ctx->mpi->size - 1)) {
        for(size_t i = 0; i < data.dim_len[1] + stencil.length; i++) {
            const size_t dest_offset = (i * dim_len_x + stencil.length) * type_size;
            void * const dest = (char *)halo.raw + dest_offset;
            const size_t len = data.dim_len[0] * type_size;
            const void * const src = (char *)data.raw - len + (i * len);
            memcpy(dest, src, len);
        }
    // Middle rank
    } else {
        for(size_t i = 0; i < data.dim_len[1] + (2 * stencil.length); i++) {
            const size_t dest_offset = (i * dim_len_x + stencil.length) * type_size;
            void * const dest = (char *)halo.raw + dest_offset;
            const size_t len = data.dim_len[0] * type_size;
            const void * const src = (char *)data.raw - len + (i * len);
            memcpy(dest, src, len);
        }
    }
    return halo;
}
