// SPDX-License-Identifier: 0BSD
/*!
 * @file
 * @license{
 * BSD Zero Clause License
 *
 * Copyright (c) 2022 by Raphael Bach
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 * }
 */
/*==============================================================================
    INCLUDE
==============================================================================*/
// Own header
#include "fmpi_reduce.h"
// C Standard Library
#include <assert.h>
#include <stdbool.h>
#include <stdio.h>  // fprintf()
#include <stdlib.h> // NULL, size_t, free(), malloc()
// Internal
#include "internal/fmpi_ctx.h"
#include "internal/fmpi_futhark.h"
#include "internal/fmpi_futhark_entry.h"
#include "internal/fmpi_mpi.h"
#include "internal/generic/fmpi_generic.h"
#include "internal/generic/fmpi_type_generic.h"
/*==============================================================================
    PUBLIC FUNCTION DEFINITION
==============================================================================*/
#define FMPI_REDUCE_DEFINITION(T) \
T fmpi_local_reduce_prod_##T( \
    const struct fmpi_ctx * const ctx, const T * const array, const size_t length \
){ \
    assert(ctx != NULL); \
    assert(array != NULL); \
    struct futhark_##T##_1d * data = fmpi_futhark_new_1d_##T(ctx->fut, array, length, 0, 0);\
    T result = FMPI_TYPE_DEFAULT_##T; \
    const int err = futhark_entry_reduce_prod_##T(ctx->fut->ctx, &result, data); \
    if(err != 0) { \
        fmpi_futhark_check_error(ctx->fut, CPL_STRINGIFY(futhark_entry_reduce_prod_##T)); \
    } \
    fmpi_futhark_sync(ctx->fut); \
    fmpi_futhark_free_1d_##T(ctx->fut, data); \
    return result; \
}\
T fmpi_reduce_prod_##T(const struct fmpi_ctx * const ctx, const T * const array) \
{ \
    assert(ctx != NULL); \
    assert(array != NULL); \
    T result = FMPI_TYPE_DEFAULT_##T; \
    /*! @cast MPI uses `int` for `size`. */ \
    /* Casting `int` to `size_t` should be safe if the value isn't negative. */ \
    /* This is asserted in`fmpi_init()`.*/ \
    T * fact_global = malloc((size_t)ctx->mpi->size * sizeof(*fact_global)); \
    if(fact_global == NULL) { \
        fprintf(stderr, "malloc(fact_global) failed!\n"); \
    } \
    int err_id = MPI_Gather(array, 1, FMPI_TYPE_MPI_##T, fact_global, 1, FMPI_TYPE_MPI_##T, ctx->mpi->root, ctx->mpi->world); \
    if(fmpi_mpi_check_error(ctx->mpi, err_id, "MPI_Gather()") == true) { \
        free(fact_global); \
    } \
    if(fmpi_mpi_is_root(ctx->mpi)) { \
        result = fmpi_local_reduce_prod_##T(ctx, fact_global, (size_t)ctx->mpi->size); \
    } \
    free(fact_global); \
    return result; \
}

FMPI_DEFINE_FUNCS_T(FMPI_REDUCE_DEFINITION, FMPI_REDUCE_PROD_TYPES)
